-----------------------------------------------------------------
-- Alloy4DP -- Alloy for Declarative Processes
-- Copyright (C) 2012, 2013  Yoann Laurent (yoann.laurent@lip6.fr)

-- This program is free software: you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation, either version 3 of the
-- License, or (at your option) any later version.

-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see
-- <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------

module declarative

open util/ordering[State] as S
open util/integer
open util/boolean
open util/relation


/************************************************************************
Signatures and functions to represent the process model
************************************************************************/

abstract sig Object {}

abstract sig Orga 						extends Object {}
abstract sig Workflow 				extends Object {}

abstract sig Activity 					extends Workflow {}

abstract sig Resource 				extends Orga {}
abstract sig Agent 					extends Resource {}
abstract sig Role 						extends Orga {}

abstract sig Data 						extends Orga {}
abstract sig Artifact 					extends Data {}
abstract sig Model					extends Artifact {}

fun use_resource 		: Activity -> Resource 		{ Info.use_resource_i }
fun super 					: Activity -> Activity 			{ Info.super_i }
fun role 					: Activity -> Role 				{ Info.role_i }
fun performer 			: Activity -> Agent 				{ Info.performer_i }

abstract sig Info {
	use_resource_i 		: Activity -> Resource	,
	super_i 					: Activity -> Activity 	,
	role_i 					: Activity -> Role 		,
	performer_i 			: Activity -> Agent
} 


/************************************************************************
State, Traces
************************************************************************/

sig State {
	exec : set Activity,  
	time : Int,
	data : set Data, 
	running : Status
}

abstract sig Status {}
one sig Running, Finished extends Status {}

fun a_nexts[s: State]: set State { s+S/nexts[s] }
fun a_prevs[s: State]: set State { s+S/prevs[s] }
fun State.c_prev : State { this = first => first else this.prev}
fun State.c_next : State { this = last => last else this.next}
pred State.setExec[n:set Activity] { this.exec = n }
pred State.setTime[t:Int] { this.time = t }
pred State.setData[d:set Data] { this.data = d }
pred State.setRunning[s:Status] { this.running = s }
pred State.isExec[n:Activity] { n in this.exec }
pred State.isRunning[] { this.running = Running }
pred State.data_exist[d:Data] { d in this.data }
pred State.data_not_exist[d:Data] { not (d in this.data) }

pred setState[s:State, e:Activity] {
	s.setExec[e] 
	s.setTime[s.c_prev.time] 
	s.setData[s.c_prev.data] 
	s.setRunning[Running]
}
pred setState[s:State, e: Activity, d:Data] {
	s.setExec[e] 
	s.setTime[s.c_prev.time] 
	s.setData[s.c_prev.data ++ d] 
	s.setRunning[Running]
}
pred setState[s:State, e: Activity, t: Int, d:Data] {
	s.setExec[e] 
	s.setTime[t] 
	s.setData[d] 
	s.setRunning[Running]
}
pred setState[s:State, e: Activity, t: Int, d:Data, st:Status] {
	s.setExec[e] 
	s.setTime[t] 
	s.setData[d] 
	s.setRunning[st]
}


/** Traces Pattern */
pred traces[from:State] {
	all s: from.a_nexts - last | let s' = s.next | {
		s.running = Running and doTransition[s,s'] implies { 
			setExecuting[s,s']
			activity_transition[s,s'] 
			update_time[s,s']
		} else {
			// stuttering the last state
			endLoop[s,s']
		}
	}
}

// set the initial state
fact initialState {
	setState[first, none, 0, none]
}

// proper termination of the trace
fact properTermination {
	// end
	some s:State | s.running = Finished
	// end properly
	all s:State | s.running = Finished implies s.exec = none
	last.exec = none
}

/* constraint the possibility of transition*/
pred activity_transition[s,s':State] { 
	// only one activity can finish
	(one n:Activity |  activity_finish[s,n] and blockOtherNodes[s,s',n])   or
	// only one activity can start
	(one n:Activity |  activity_start[s,n] and blockOtherNodes[s,s',n]) 	or
	// no activity start, i.e. only the time can increase
	(no a:Activity |  (activity_start[s,a] or activity_finish[s,a])) 
}

// update the time
pred update_time[s,s':State] {
	gte[s'.time, s.time]
}

pred blockOtherNodes[s,s':State,n:Activity] {
	all a:Activity-n | {a in s.exec implies a in s'.exec}
	all a:Activity-n | {not a in s.exec implies not a in s'.exec}
}

pred doTransition[s,s':State] 	{ s'.exec != s.exec }
pred setExecuting[s,s':State] 	{ s'.running = Running }
pred endLoop[s,s':State] 			{ setState[s', s.exec, s.time, s.data, Finished] }

pred activity_start[s:State, n:Activity]  { not s.isExec[n] and s.c_next.isExec[n] }
pred activity_finish[s:State, n:Activity] { s.isExec[n] and not s.c_next.isExec[n] }




/************************************************************************
Process Constraints
************************************************************************/

// NOTE: predicate starting by "_" are 'private', 
//			  they are useful only to define the process constraints

/**---------------------- helper function ------------------*/
// return all the State between 'start' and 'end' on which the 'activities' start
fun getStartActivation[start,end:State, activities: set Activity] : State -> Activity
	{ { s: State, a:activities | s in a_nexts[start] and s.c_next in a_prevs[end] and activity_start[s,a] } }
// return all the State on which the 'activities' start
fun getStartActivation[activities:Activity] : State -> Activity
	{ getStartActivation[first,last,activities] }

// return all the State between 'start' and 'end' on which the activity 'a' finish
fun getFinishActivation[start,end:State, activities:Activity] : State -> Activity
	{ { s: State, a:activities | s in a_nexts[start] and s.c_next in a_prevs[end] and activity_finish[s,a] } }
// return all the State on which the activity 'a' start
fun getFinishActivation[activities:Activity] : State -> Activity
	{ getFinishActivation[first,last,activities] }
// return the tuple of State representing all the start and end of the activation of the activity 'a' between 'start' and 'end'
fun getActivation[start,end:State, activities: Activity] : State -> State -> Activity
	{ { 
		disj s1,s2 : State, a:activities | 	s1 in a_nexts[start] and s1.next in a_prevs[end] and activity_start[s1,a] and
							 		s2 in a_nexts[start] and s2.next in a_prevs[end] and activity_finish[s2,a] and
									s2 in s1.nexts and
									no getStartActivation[s1.next, s2,a]	 // (no activation between the two State)
	} }
// return the tuple of State representing all the start and end of the activation of the activity 'a' between 'start' and 'end'
fun getActivation[activities: set Activity] : State -> State -> Activity {
	{ getActivation[first, last, activities] }
}
// return all the State on which an Activity starts
fun getStartActivationAll[start,end:State] : State -> Activity
	{ { s: State, a:Activity | s in a_nexts[start] and s.c_next in a_prevs[end] and activity_start[s,a] } }
/**--------------- helper function time related ---------------*/
// return the maximum State of time 't'
fun maxStateTime[t:Int] : State { { S/max[{ s:State | s.time <= t }] } }
// return the minimum State of time 't'
fun minStateTime[t:Int] : State { { S/min[{ s:State | s.time >= t }] } }
//return the State with a "time" between start and end 
fun rangeStateTime[start,end:Int] : State 
{  { s : State | s in a_nexts[minStateTime[start]] and s in a_prevs[maxStateTime[end]] } }
// return the maximum time allowed
fun maxTime : Int { integer/max }
// return the minimum time allowed
fun minTime : Int { integer/min }
/**-----------------------------------------------*/


/** composition of activity */
// the 'childs' activity can start and terminate only if 'super' is executing
pred childOf[super, childs: set Activity] {
	all s:State, child:childs| (activity_start[s,child] or  activity_finish[s,child])
		implies
			s.isExec[super]
}

/** AtomicActivity */
// the 'activities' execution must be atomic
pred atomic[activities: set Activity] {
	all s:State, a:activities | activity_start[s,a] implies { activity_finish[s.c_next,a] }
}

/***********************
* Existence
***********************/

// activity 'n' happens between 'start' and 'end', a number of times between 'mi' and 'ma'
pred _existence[start,end:State, n:Activity, mi,ma:Int]  {	
	#{getStartActivation[start,end,n]} >= mi and #{getStartActivation[start,end,n]} <= ma 
}
// activity 'n' happens a number of times between 'mi' and 'ma'
pred existence[n:Activity, mi,ma:Int] {	
	_existence[first,last,n,mi,ma]
}

// activity 'a' happens between 'start' and 'end'
pred _existence[start, end:State, a:Activity] {
	some s:State | s.isExec[a] and s in a_nexts[start] and s in a_prevs[end]
}
// activity 'a' happens between
pred existence[a: Activity] {	
	_existence[first,last, a] 
}

// activity 'a' never happens between 'start' and 'end'
pred _abscence[start, end:State, a:Activity] {
	no s:State | s.isExec[a] and s in a_nexts[start] and s in a_prevs[end]
}
// activity 'a' never happens
pred abscence[a:Activity] {	
	_abscence[first,last, a] 
}

// activity 'a' terminate between 'start' and 'end'
pred _existence_finish[from,to:State, a:Activity] {
	some s:State | s in a_nexts[from] and s.c_next in a_prevs[to] and activity_finish[s,a]
}

// activity 'a' start between 'start' and 'end'
pred _existence_start[from,to:State, a:Activity] {
	some s:State | s in a_nexts[from] and s.c_next in a_prevs[to] and activity_start[s,a]
}

// activity 'a' start and terminate between 'start' and 'end'
pred _existence_start_finish[from,to:State, a:Activity] {
	_existence_start[from,to,a] 
	_existence_finish[from,to,a] 
}

// all the execution of 'a' must be between 'start' and 'end'
pred _exclusive_existence[start, end:State, a:Activity] {
	all s:State | s.isExec[a] implies {
		s in a_nexts[start] and s in a_prevs[end]
	}
}
// all the execution of 'a' must not be between 'start' and 'end'
pred _exclusive_abscence[start, end:State, a:Activity] {
	all s:State | s.isExec[a] implies {
		not (s in a_nexts[start] and s in a_prevs[end])
	}
}

/** existence within a set */

// some of 'activities' must be executed an arbitrary number of time
pred some_existence[activities : set Activity] {
	some a:activities | existence[a]
}

// more than 'n' number of 'activities' must be executed an arbitrary number of time
pred more_existence[activities : set Activity, n:Int] {
	#{ a:activities | existence[a] } > n
}

// less than 'n' number of 'activities' must be executed an arbitrary number of time
pred less_existence[activities : set Activity, n:Int] {
	#{ a:activities | existence[a] } < n
}

// only one of 'activities' must be executed; other must not exist
pred one_existence[activities : set Activity] {
	one a:activities | existence[a]
}

// only 'n' number of 'activities' must be executed; other must not exist
pred exactly_existence[activities: set Activity, n:Int] {
	#{ a:activities | existence[a] } = n
}

/** initial / final */

// 'a' must be the first activity executed
pred initial[a:Activity] {
	a in State1.exec
}

// 'a' must be the last activity terminated
pred final[a:Activity] {
	some s:State | activity_finish[s,a] and not s.c_next.next.isRunning[]			  		
}


/** number of activations */

// 'a' and 'b' must be executed the same number of times
pred executed_same_number[a,b: set Activity]{
	#{ getStartActivation[a] } = 	#{ getStartActivation[b] }
}

// 'a' and 'b' must not be executed the same number of times
pred negation_executed_same_number[a,b: set Activity]{
	not executed_same_number[a,b]
}

// 'a' must be executed less times than 'b'
pred executed_lt_number[a,b: set Activity]{
	#{ getStartActivation[a] } < 	#{ getStartActivation[b] }
}

// 'a' must be executed more times than 'b'
pred executed_gt_number[a,b: set Activity]{
	#{ getStartActivation[a] } > 	#{ getStartActivation[b] }
}


/** number of activation for the whole Activities */

// the total number of Activity executed is equal to 'total'
pred exactly_activity_executed[total:Int] {
	#{ getStartActivationAll[first,last] } = total
}
// the total number of Activity executed is more than 'total'
pred more_activity_executed[total:Int] {
	#{ getStartActivationAll[first,last] } > total
}
// the total number of Activity executed is less than 'total'
pred less_activity_executed[total:Int] {
	#{ getStartActivationAll[first,last] } < total
}


/***********************
* Relation Existence
***********************/

pred existence_response[a,b:Activity] {	
	existence[a] implies existence[b]
}

pred negation_existence_response[a,b:Activity] {	
	existence[a] implies abscence[b]
}

pred co_existence[a,b:Activity] {	
	existence[a] iff existence[b]
}

pred negation_co_existence[a,b:Activity] {	
	negation_existence_response[a,b]
	and
	negation_existence_response[b,a]
}

/***********************
* Parallal exclusion
***********************/

// the acitivites are executed in parallal on some state during the execution
pred parallal[activities: set Activity] {
	some s:State | { all a:activities | s.isExec[a] }
}

// maximum number of activities which can be executed simultanously
pred max_parallal[ i : Int ] {
	all s:State | #{  a: s.exec | no a.super } <= i
}

// maximum number of activities which can be executed simultanously between the 'activities'
pred max_parallal[ activities: set Activity, i : Int ] {
	all s:State | #{  a:activities | s.isExec[a] } <= i
}

pred exclusion[activities: set Activity] {	
	all s:State | some s.exec implies {
		(lone a:activities | s.isExec[a])
	}
}


/***********************
* Relation Order
***********************/


/** Response */
pred response[a,b:Activity] {	
	all s:State | activity_finish[s,a] 
		implies
			_existence_start[s.c_next, last,b]
}
pred chain_response[a,b:Activity] {	
	all s:State | activity_finish[s,a] 
		implies
			_existence_start[s.c_next, s.c_next.next,b]
}
pred negation_response[a,b:Activity] {	
	all s:State | activity_finish[s,a] 
		implies
			not _existence_start[s.c_next,last,b] 
}
pred negation_chain_response[a,b:Activity] {	
	all s:State | activity_finish[s,a] 
		implies
			not _existence_start[s.c_next, s.c_next.next,b]
}

pred alternate_response[a,b:Activity] {
	response[a,b]
	all disj s1,s2:State | activity_finish[s1,a] and activity_start[s2,a] and s2 in s1.nexts
		implies {
			_existence_start_finish[s1.next, s2,b]
		}
}
pred negation_alternate_response[a,b:Activity] {
	negation_response[a,b]
	all disj s1,s2:State | activity_finish[s1,a] and activity_start[s2,a] and s2 in s1.nexts
		implies {
			not _existence_start_finish[s1.next, s2,b]
		}
}

/** Precedence */
pred precedence[a,b:Activity] {	
	all s:State | activity_start[s,a] 
		implies
			_existence_finish[first,s,b]
}
pred chain_precedence[a,b:Activity] {	
	all s:State | activity_start[s,a]  
		implies		
			_existence_finish[s.c_prev,s,b]						
}
pred negation_precedence[a,b:Activity] {	
	all s:State | activity_start[s,a] 
		implies
			not _existence_finish[first,s,b] 
}
pred negation_chain_precedence[a,b:Activity] {	
	all s:State | activity_start[s,a]  
		implies		
			not _existence_finish[s.c_prev,s,b]						
}

pred alternate_precedence[a,b:Activity] {
	precedence[a,b]
	all disj s1,s2:State | activity_start[s1,a] and activity_finish[s2,a] and s2 in s1.prevs
		implies {
			_existence_start_finish[s2.next, s1,b]
		}
}

pred negation_alternate_precedence[a,b:Activity] {
	negation_precedence[a,b]
	all disj s1,s2:State | activity_start[s1,a] and activity_finish[s2,a] and s2 in s1.prevs
		implies {
			not _existence_start_finish[s2.next, s1,b]
		}
}

/** Succession */
pred succession[a,b:Activity] {	
	response[a,b] and precedence[b,a]
}
pred chain_succession[a,b:Activity] {
	chain_response[a,b] and chain_precedence[b,a]
}
pred negation_succession[a,b:Activity] {	
	not succession[a,b]
}
pred negation_chain_succession[a,b:Activity] {
	not chain_succession[a,b]
}
pred alternate_succession[a,b:Activity] {
	alternate_response[a,b] and alternate_precedence[b,a]
}
pred negation_alternate_succession[a,b:Activity] {
	negation_alternate_response[a,b] and negation_alternate_precedence[b,a]
}


/***********************
* N-ary relation
***********************/
// 'be' is always executed between an execution of 'a' and 'b'
pred between[be,a,b:Activity] {
	all s : State | s.isExec[be] implies {
			_existence_start_finish[first, s.c_prev, a]
			and 
			_existence_start_finish[s.c_next, last, b]
	}
}
// 'be' is never executed between an execution of 'a' and 'b'
pred negation_between[be,a,b:Activity] {
	all s : State | s.isExec[be] implies {
			not _existence_start_finish[first, s.c_prev, a]
			and 
			not _existence_start_finish[s.c_next, last, b]
	}
}
//  'be' are executed between all the execution of 'a' and 'b'
pred strong_between[be,a,b:Activity] {
	all s : State | s.isExec[be] implies {
			_existence_start_finish[first, s.c_prev, a] and _exclusive_existence[first, s.c_prev, a]
			and 
			_existence_start_finish[s.c_next, last, b] and _exclusive_existence[s.c_next, last, b]
	}
}

/***********************
* Data flow
***********************/

// all/some/one/lone/no data 'd' must be available before activity 'a' start 

pred all_data_in[a:Activity, d: set Data] 
{ all s:State | activity_start[s,a] implies s.data_exist[d] and  s.c_next.data = s.data }

pred some_data_in[a:Activity, d: set Data] 
{ all s:State | activity_start[s,a] implies (some dt:d | dt in s.data) and  s.c_next.data = s.data }

pred one_data_in[a:Activity, d: set Data] 
{ all s:State | activity_start[s,a] implies (one dt:d | dt in s.data) and  s.c_next.data = s.data }

pred lone_data_in[a:Activity, d: set Data] 
{ all s:State | activity_start[s,a] implies (lone dt:d | dt in s.data) and  s.c_next.data = s.data }

pred no_data_in[a:Activity, d: set Data] 
{ all s:State | activity_start[s,a] implies (no dt:d | dt in s.data) and  s.c_next.data = s.data }

// all/some/one/lone/no data 'd' must be produced at the end of the activity 'a' 

pred all_data_out[a:Activity, d: set Data] 
{ all s:State | activity_finish[s,a] implies s.c_next.data = s.data ++ d }

pred some_data_out[a:Activity, d: set Data] 
{ all s:State | activity_finish[s,a] implies s.data in s.c_next.data and (some dt:d | dt in s.c_next.data) }

pred one_data_out[a:Activity, d: set Data] 
{ all s:State | activity_finish[s,a] implies s.data in s.c_next.data and (one dt:d | dt in s.c_next.data) }

pred lone_data_out[a:Activity, d: set Data] 
{ all s:State | activity_finish[s,a] implies s.data in s.c_next.data and (lone dt:d | dt in s.c_next.data) }

pred no_data_out[a:Activity, d: set Data] 
{ all s:State | activity_finish[s,a] implies s.data in s.c_next.data and (no dt:d | dt in s.c_next.data) }



pred all_data_in_out[a:Activity, i: set Data, o: set Data] 
{ all_data_in[a,i] and all_data_out[a,o] }


// all/some/one/lone/no data 'd' must be available during the process execution

pred d_all_presence[d: set Data] {
	all dt : d | some s:State |  s.data_exist[dt]
}
pred d_some_presence[d: set Data] {
	some dt : d | some s:State |  s.data_exist[dt]
}
pred d_one_presence[d: set Data] {
	one dt : d | some s:State |  s.data_exist[dt]
}
pred d_lone_presence[d: set Data] {
	lone dt : d | some s:State |  s.data_exist[dt]
}
pred d_no_presence[d: set Data] {
	no dt : d | some s:State |  s.data_exist[dt]
}


/***********************
* Timed
***********************/

pred t_existence_start[a:Activity,ti,tf:Int] {	
	some s : State | 
		activity_start[s,a] and
		s.c_next.time >= ti and
		s.c_next.time <= tf 
}
pred t_existence_finish[n:Activity,ti,tf:Int] {	
	some s : State | 
		activity_finish[s,n] and
		s.c_next.time >= ti and
		s.c_next.time <= tf 
}

// the activity 'a' must be executed between 'durationMin' and 'durationMax' time unit
pred t_duration[a:Activity,durationMin:Int, durationMax:Int] {	
	all s1,s2 : State | activity_start [s1,a] and activity_finish[s2,a] and s2 in a_nexts[s1.next]
		implies {
			t_existence_start[a,s1.next.time,s1.next.time] and
			t_existence_finish[a, s1.next.time.add[durationMin],s1.next.time.add[durationMax]]
		}
}

// the activity 'a' must be executed in 'duration' time unit
pred t_duration[a:Activity,duration:Int] {	
	t_duration[a,duration,duration]
}

// the activity 'a' must be executed between the time unit 'start' and 'finish'
pred t_existence[a:Activity,start,finish:Int] {	
	t_existence_start[a,start,finish] and 
	t_existence_finish[a,start,finish]
}

// the activity 'a' must be not be executed between the time unit 'start' and 'finish'
pred t_abscence[a:Activity, start,end:Int] {
	not t_existence[a,start,end]
}

// the activity 'a' must be executed ONLY between the time unit 'start' and 'finish'
pred t_exclusive_allowance[a:Activity,start,end:Int] {
	t_existence[a,start,end] 
	and 
	#{s : rangeStateTime[start,end] | activity_start[s,a]} = #{s : rangeStateTime[start,end] | activity_finish[s,a]}
	and
	t_abscence[a,	 minTime, minStateTime[start].c_prev.time]
	and
	t_abscence[a,	 maxStateTime[end].c_next.time,	maxTime]	
}

// the data 'd' must be available before 'min' time unit
pred t_data_available[d:Data, min:Int] {
	some s : State | s.data_exist[d] and lte[s.time,min]
}


/***********************
* Resources
***********************/

// resource 'r' can be used only by 'nb' activities at the same times
pred limited_use_resource[r:Resource, nb : Int] {
	all s:State |  #{ a : s.exec | r in a.use_resource } <= nb
}

// when the activity 'a' are executing, only 'a' can use the resource 'r'
pred exclusice_use_resource[r:Resource,  a:Activity] {
	no s:State, other:(Activity-a) | r in other.use_resource and s.isExec[other] and s in { s:State | s.isExec[a] }
}




/***********************
* Helper to declare the trace
***********************/
fun State0: State  { S/first }
fun State1: State         { State0.next } 
fun State2: State         { State1.next } 
fun State3: State         { State2.next } 
fun State4: State         { State3.next } 
fun State5: State         { State4.next } 
fun State6: State         { State5.next } 
fun State7: State         { State6.next } 
fun State8: State         { State7.next } 
fun State9: State         { State8.next } 
fun State10: State         { State9.next } 
fun State11: State         { State10.next } 
fun State12: State         { State11.next } 
fun State13: State         { State12.next } 
fun State14: State         { State13.next } 
fun State15: State         { State14.next } 
fun State16: State         { State15.next } 
fun State17: State         { State16.next } 
fun State18: State         { State17.next } 
fun State19: State         { State18.next } 
fun State20: State         { State19.next } 
fun State21: State         { State20.next } 
fun State22: State         { State21.next } 
fun State23: State         { State22.next } 
fun State24: State         { State23.next } 
fun State25: State         { State24.next } 
fun State26: State         { State25.next } 
fun State27: State         { State26.next } 
fun State28: State         { State27.next } 
fun State29: State         { State28.next } 
fun State30: State         { State29.next } 
fun State31: State         { State30.next } 
fun State32: State         { State31.next } 
fun State33: State         { State32.next } 
fun State34: State         { State33.next } 
fun State35: State         { State34.next } 
fun State36: State         { State35.next } 
fun State37: State         { State36.next } 
fun State38: State         { State37.next } 
fun State39: State         { State38.next } 
fun State40: State         { State39.next } 
fun State41: State         { State40.next } 
fun State42: State         { State41.next } 
fun State43: State         { State42.next } 
fun State44: State         { State43.next } 
fun State45: State         { State44.next } 
fun State46: State         { State45.next } 
fun State47: State         { State46.next } 
fun State48: State         { State47.next } 
fun State49: State         { State48.next } 
fun State50: State         { State49.next } 
fun State51: State         { State50.next } 
fun State52: State         { State51.next } 
fun State53: State         { State52.next } 
fun State54: State         { State53.next } 
fun State55: State         { State54.next } 
fun State56: State         { State55.next } 
fun State57: State         { State56.next } 
fun State58: State         { State57.next } 
fun State59: State         { State58.next } 
fun State60: State         { State59.next } 
fun State61: State         { State60.next } 
fun State62: State         { State61.next } 
fun State63: State         { State62.next } 
fun State64: State         { State63.next } 
fun State65: State         { State64.next } 
fun State66: State         { State65.next } 
fun State67: State         { State66.next } 
fun State68: State         { State67.next } 
fun State69: State         { State68.next } 
fun State70: State         { State69.next } 
fun State71: State         { State70.next } 
fun State72: State         { State71.next } 
fun State73: State         { State72.next } 
fun State74: State         { State73.next } 
fun State75: State         { State74.next } 
fun State76: State         { State75.next } 
fun State77: State         { State76.next } 
fun State78: State         { State77.next } 
fun State79: State         { State78.next } 
fun State80: State         { State79.next } 
fun State81: State         { State80.next } 
fun State82: State         { State81.next } 
fun State83: State         { State82.next } 
fun State84: State         { State83.next } 
fun State85: State         { State84.next } 
fun State86: State         { State85.next } 
fun State87: State         { State86.next } 
fun State88: State         { State87.next } 
fun State89: State         { State88.next } 
fun State90: State         { State89.next } 
fun State91: State         { State90.next } 
fun State92: State         { State91.next } 
fun State93: State         { State92.next } 
fun State94: State         { State93.next } 
fun State95: State         { State94.next } 
fun State96: State         { State95.next } 
fun State97: State         { State96.next } 
fun State98: State         { State97.next } 
fun State99: State         { State98.next } 
fun State100: State         { State99.next } 
fun State101: State         { State100.next } 
fun State102: State         { State101.next } 
fun State103: State         { State102.next } 
fun State104: State         { State103.next } 
fun State105: State         { State104.next } 
fun State106: State         { State105.next } 
fun State107: State         { State106.next } 
fun State108: State         { State107.next } 
fun State109: State         { State108.next } 
fun State110: State         { State109.next } 
fun State111: State         { State110.next } 
fun State112: State         { State111.next } 
fun State113: State         { State112.next } 
fun State114: State         { State113.next } 
fun State115: State         { State114.next } 
fun State116: State         { State115.next } 
fun State117: State         { State116.next } 
fun State118: State         { State117.next } 
fun State119: State         { State118.next } 
fun State120: State         { State119.next } 
fun State121: State         { State120.next } 
fun State122: State         { State121.next } 
fun State123: State         { State122.next } 
fun State124: State         { State123.next } 
fun State125: State         { State124.next } 
fun State126: State         { State125.next } 
fun State127: State         { State126.next } 
fun State128: State         { State127.next } 
fun State129: State         { State128.next } 
fun State130: State         { State129.next } 
fun State131: State         { State130.next } 
fun State132: State         { State131.next } 
fun State133: State         { State132.next } 
fun State134: State         { State133.next } 
fun State135: State         { State134.next } 
fun State136: State         { State135.next } 
fun State137: State         { State136.next } 
fun State138: State         { State137.next } 
fun State139: State         { State138.next } 
fun State140: State         { State139.next } 
fun State141: State         { State140.next } 
fun State142: State         { State141.next } 
fun State143: State         { State142.next } 
fun State144: State         { State143.next } 
fun State145: State         { State144.next } 
fun State146: State         { State145.next } 
fun State147: State         { State146.next } 
fun State148: State         { State147.next } 
fun State149: State         { State148.next } 
fun State150: State         { State149.next } 
fun State151: State         { State150.next } 
fun State152: State         { State151.next } 
fun State153: State         { State152.next } 
fun State154: State         { State153.next } 
fun State155: State         { State154.next } 
fun State156: State         { State155.next } 
fun State157: State         { State156.next } 
fun State158: State         { State157.next } 
fun State159: State         { State158.next } 
fun State160: State         { State159.next } 
fun State161: State         { State160.next } 
fun State162: State         { State161.next } 
fun State163: State         { State162.next } 
fun State164: State         { State163.next } 
fun State165: State         { State164.next } 
fun State166: State         { State165.next } 
fun State167: State         { State166.next } 
fun State168: State         { State167.next } 
fun State169: State         { State168.next } 
fun State170: State         { State169.next } 
fun State171: State         { State170.next } 
fun State172: State         { State171.next } 
fun State173: State         { State172.next } 
fun State174: State         { State173.next } 
fun State175: State         { State174.next } 
fun State176: State         { State175.next } 
fun State177: State         { State176.next } 
fun State178: State         { State177.next } 
fun State179: State         { State178.next } 
fun State180: State         { State179.next } 
fun State181: State         { State180.next } 
fun State182: State         { State181.next } 
fun State183: State         { State182.next } 
fun State184: State         { State183.next } 
fun State185: State         { State184.next } 
fun State186: State         { State185.next } 
fun State187: State         { State186.next } 
fun State188: State         { State187.next } 
fun State189: State         { State188.next } 
fun State190: State         { State189.next } 
fun State191: State         { State190.next } 
fun State192: State         { State191.next } 
fun State193: State         { State192.next } 
fun State194: State         { State193.next } 
fun State195: State         { State194.next } 
fun State196: State         { State195.next } 
fun State197: State         { State196.next } 
fun State198: State         { State197.next } 
fun State199: State         { State198.next } 
fun State200: State         { State199.next } 
fun State201: State         { State200.next } 
fun State202: State         { State201.next } 
fun State203: State         { State202.next } 
fun State204: State         { State203.next } 
fun State205: State         { State204.next } 
fun State206: State         { State205.next } 
fun State207: State         { State206.next } 
fun State208: State         { State207.next } 
fun State209: State         { State208.next } 
fun State210: State         { State209.next } 
fun State211: State         { State210.next } 
fun State212: State         { State211.next } 
fun State213: State         { State212.next } 
fun State214: State         { State213.next } 
fun State215: State         { State214.next } 
fun State216: State         { State215.next } 
fun State217: State         { State216.next } 
fun State218: State         { State217.next } 
fun State219: State         { State218.next } 
fun State220: State         { State219.next } 
fun State221: State         { State220.next } 
fun State222: State         { State221.next } 
fun State223: State         { State222.next } 
fun State224: State         { State223.next } 
fun State225: State         { State224.next } 
fun State226: State         { State225.next } 
fun State227: State         { State226.next } 
fun State228: State         { State227.next } 
fun State229: State         { State228.next } 
fun State230: State         { State229.next } 
fun State231: State         { State230.next } 
fun State232: State         { State231.next } 
fun State233: State         { State232.next } 
fun State234: State         { State233.next } 
fun State235: State         { State234.next } 
fun State236: State         { State235.next } 
fun State237: State         { State236.next } 
fun State238: State         { State237.next } 
fun State239: State         { State238.next } 
fun State240: State         { State239.next } 
fun State241: State         { State240.next } 
fun State242: State         { State241.next } 
fun State243: State         { State242.next } 
fun State244: State         { State243.next } 
fun State245: State         { State244.next } 
fun State246: State         { State245.next } 
fun State247: State         { State246.next } 
fun State248: State         { State247.next } 
fun State249: State         { State248.next } 
fun State250: State         { State249.next } 
fun State251: State         { State250.next } 
fun State252: State         { State251.next } 
fun State253: State         { State252.next } 
fun State254: State         { State253.next } 
fun State255: State         { State254.next } 
fun State256: State         { State255.next } 
fun State257: State         { State256.next } 
fun State258: State         { State257.next } 
fun State259: State         { State258.next } 
fun State260: State         { State259.next } 
fun State261: State         { State260.next } 
fun State262: State         { State261.next } 
fun State263: State         { State262.next } 
fun State264: State         { State263.next } 
fun State265: State         { State264.next } 
fun State266: State         { State265.next } 
fun State267: State         { State266.next } 
fun State268: State         { State267.next } 
fun State269: State         { State268.next } 
fun State270: State         { State269.next } 
fun State271: State         { State270.next } 
fun State272: State         { State271.next } 
fun State273: State         { State272.next } 
fun State274: State         { State273.next } 
fun State275: State         { State274.next } 
fun State276: State         { State275.next } 
fun State277: State         { State276.next } 
fun State278: State         { State277.next } 
fun State279: State         { State278.next } 
fun State280: State         { State279.next } 
fun State281: State         { State280.next } 
fun State282: State         { State281.next } 
fun State283: State         { State282.next } 
fun State284: State         { State283.next } 
fun State285: State         { State284.next } 
fun State286: State         { State285.next } 
fun State287: State         { State286.next } 
fun State288: State         { State287.next } 
fun State289: State         { State288.next } 
fun State290: State         { State289.next } 
fun State291: State         { State290.next } 
fun State292: State         { State291.next } 
fun State293: State         { State292.next } 
fun State294: State         { State293.next } 
fun State295: State         { State294.next } 
fun State296: State         { State295.next } 
fun State297: State         { State296.next } 
fun State298: State         { State297.next } 
fun State299: State         { State298.next } 
fun State300: State         { State299.next } 
fun State301: State         { State300.next } 













